// Copyright (C) 1999-2006
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "tcl.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "framergbtruecolor.h"
#include "fitsimage.h"
#include "util.h"
#include "ps.h"
#include "NaN.h"

FrameRGBTrueColor::FrameRGBTrueColor(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item)
  : FrameBase(i,c,item), FrameRGB(i,c,item), FrameTrue(i,c,item)
{
  for (int kk=0; kk<3; kk++)
    colormapData[kk] = NULL;
}

FrameRGBTrueColor::~FrameRGBTrueColor()
{
  for (int kk=0; kk<3; kk++)
    if (colormapData[kk])
      delete [] colormapData[kk];
}

// fillRGBImage

unsigned char* FrameRGBTrueColor::fillRGBImage(int width, int height,
					       int x0, int y0, int x1, int y1,
					     double* (FitsImage::*getMatrix)())
{
  // create 3 byte img
  unsigned char* img = new unsigned char[width*height*3];
  memset(img, 0, width*height*3);

  // special case
  if ((!view[0] || !fits[0]) &&
      (!view[1] || !fits[1]) &&
      (!view[2] || !fits[2])) {

    unsigned char* ptr = img;
    for (int jj=0; jj<height; jj++)
      for (int ii=0; ii<width; ii++) {
	*img++ = (unsigned char)bgColor->red;
	*img++ = (unsigned char)bgColor->green;
	*img++ = (unsigned char)bgColor->blue;
      }	
    return img;
  }

  char* mk = new char[width*height];
  memset(mk, 0, width*height);

  // ok, create an RGB image
  for (int kk=0; kk<3; kk++) {
    if (!view[kk] || !fits[kk])
      continue;

    if (!(mosaicCount[kk]>1))
      fillRGBImageSingle(kk, img, mk, x0, y0, x1, y1, getMatrix);
    else {
      if (mosaicFast && mosaicType[kk] == IRAF)
	fillRGBImageMosaicFast(kk, img, mk, x0, y0, x1, y1, getMatrix);
      else
	fillRGBImageMosaic(kk, img, mk, x0, y0, x1, y1, getMatrix);
    }
  }

  // now, apply mask
  unsigned char* ptr = img;
  char* mkptr = mk;
  for (int jj=y0; jj<y1; jj++)
    for (int ii=x0; ii<x1; ii++, ptr+=3, mkptr++)
      if (!*mkptr) {
	*(ptr  ) = (unsigned char)bgColor->red;
	*(ptr+1) = (unsigned char)bgColor->green;
	*(ptr+2) = (unsigned char)bgColor->blue;
      }

  // clean up
  if (mk)
    delete [] mk;

  return img;
}

void FrameRGBTrueColor::fillRGBImageSingle(int kk,
					   unsigned char* img, char* mk,
					   int x0, int y0, int x1, int y1,
					   double* (FitsImage::*getMatrix)())
{
  double* mm = (cfits[kk]->*getMatrix)();
  int* params = cfits[kk]->getDataParams(frScale[kk].scanMode());
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  int length = colorScale[kk]->size() - 1;
  const unsigned char* table = colorScale[kk]->psColors();

  double ll = cfits[kk]->getLowDouble();
  double hh = cfits[kk]->getHighDouble();
  double diff = hh - ll;

  unsigned char* ptr = img;
  char* mkptr = mk;
  for (int jj=y0; jj<y1; jj++) {
    for (int ii=x0; ii<x1; ii++, ptr+=3, mkptr++) {

      double xx = ii*mm[0] + jj*mm[3] + mm[6];
      double yy = ii*mm[1] + jj*mm[4] + mm[7];

      if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	double value = cfits[kk]->getValueDouble(((int)yy)*srcw + (int)xx);
	int mask =0;
	if (cmask[kk])
	  mask = cmask[kk]->getValueMask(((int)yy)*srcw + (int)xx);
	  
	if (!mask) {
	  if (!isnand(value)) {
	    if (value <= ll)
	      *(ptr+kk) = *table;
	    else if (value >= hh)
	      *(ptr+kk) = *(table+length);
	    else
	      *(ptr+kk) = *(table+((int)(((value - ll)/diff * length) +.5)));

	    // set mask
	    *mkptr = 1;
	  }
	}
      }
    }
  }
}

void FrameRGBTrueColor::fillRGBImageMosaic(int kk, 
					   unsigned char* img, char* mk,
					   int x0, int y0, int x1, int y1,
					   double* (FitsImage::*getMatrix)())
{
  int length = colorScale[kk]->size() - 1;
  const unsigned char* table = colorScale[kk]->psColors();

  unsigned char* dest = img;
  char* mkptr = mk;
  for (int jj=y0; jj<y1; jj++) {
    for (int ii=x0; ii<x1; ii++, dest+=3, mkptr++) {
      FitsImage* sptr = cfits[kk];
      FitsImage* mptr = cmask[kk];

      while (sptr) {
	double* mm = (sptr->*getMatrix)();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale[kk].scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  double value = sptr->getValueDouble(((int)yy)*srcw + (int)xx);
	  int mask =0;
	  if (mptr)
	    mask = mptr->getValueMask(((int)yy)*srcw + (int)xx);
	  
	  if (!mask) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll)
		*(dest+kk) = *table;
	      else if (value >= hh)
		*(dest+kk) = *(table+length);
	      else
		*(dest+kk) = 
		  *(table+((int)(((value - ll)/diff * length) +.5)));

	      // set mask
	      *mkptr = 1;
	    }
	  }

	  break;
	}
	else {
	  sptr = sptr->nextMosaic();
	  if (mptr)
	    mptr = mptr->nextMosaic();
	}
      }
    }
  }
}

void FrameRGBTrueColor::fillRGBImageMosaicFast(int kk, 
					       unsigned char* img, char* mk,
					       int x0, int y0, int x1, int y1,
					     double* (FitsImage::*getMatrix)())
{
  int length = colorScale[kk]->size() - 1;
  const unsigned char* table = colorScale[kk]->psColors();

  FitsImage* sptr = currentFits;
  FitsImage* mptr = currentMask;
  int found = 0;

  unsigned char* dest = img;
  char* mkptr = mk;

  for (int jj=y0; jj<y1; jj++) {
    for (int ii=x0; ii<x1; ii++, dest+=3, mkptr++) {

      while (sptr) {
	double* mm = (sptr->*getMatrix)();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale[kk].scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  found = 1;

	  double value = sptr->getValueDouble(((int)yy)*srcw + (int)xx);
	  int mask =0;
	  if (mptr)
	    mask = mptr->getValueMask(((int)yy)*srcw + (int)xx);
	  
	  if (!mask) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll)
		*(dest+kk) = *table;
	      else if (value >= hh)
		*(dest+kk) = *(table+length);
	      else
		*(dest+kk) = 
		  *(table+((int)(((value - ll)/diff * length)+.5)));

	      // set mask
	      *mkptr = 1;
	    }
	  }

	  break;
	}
	else {
	  if (found) {
	    found = 0;
	    sptr = currentFits;
	    mptr = currentMask;
	  }
	  else {
	    sptr = sptr->nextMosaic();
	    if (mptr)
	      mptr = mptr->nextMosaic();
	  }
	}
      }

      if (!sptr) {
	sptr = currentFits;
	mptr = currentMask;
      }
    }
  }
}

// colormap

void FrameRGBTrueColor::colormapCmd(float rb, float gb, float bb, 
				    float rc, float gc, float bc, int i,
				    unsigned char* cells, int cnt)
{
  // first check for change
  if (bias[0] == rb && bias[1] == gb && bias[2] == bb && 
      contrast[0] == rc && contrast[1] == gc && contrast[2] == bc &&
      invert == i && colorCells)
    return;

  // we got a change
  bias[0] = rb;
  bias[1] = gb;
  bias[2] = bb;
  contrast[0] = rc;
  contrast[1] = gc;
  contrast[2] = bc;
  invert = i;

  updateColorCells(cells, cnt);
  updateColorScale();

  if (mosaicCount[0] || mosaicCount[1] || mosaicCount[2]) {
    update(BASE); // always update
    skipUpdate = 0;
  }
  else {
    if (!skipUpdate) {
      update(BASE); // update once
      skipUpdate = 1;
    }      
  }
}

void FrameRGBTrueColor::colormapBeginCmd(const Vector& w)
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  // we need some fits data
  // we assume the colorScale length will not change during motion calls
  if (!fits[0] && !fits[1] && !fits[2])
    return;

  int width = options->width;
  int height = options->height;
  cmapOffset = Vector();
  if (cmapArea) {
    width = height = cmapArea;

    switch (cmapMode) {
    case CENTER:
      cmapOffset = (Vector(options->width,options->height) - 
		    Vector(width,height))/2;
      break;
    case CLICK: 
      {
	Vector where = w * canvasToWidget;
	if (where[0] >= 0 && where[0] < options->width &&
	    where[1] >= 0 && where[1] < options->height)
	  cmapOffset = where - Vector(width,height)/2;
	else
	  cmapOffset = (Vector(options->width,options->height) - 
			Vector(width,height))/2;
      }
      break;
    }

    if (cmapOffset[0] < 0) {
      width += (int)cmapOffset[0];
      cmapOffset[0] = 0;
    }
    if (cmapOffset[0]+width >= options->width)
      width = options->width - (int)cmapOffset[0];

    if (cmapOffset[1] < 0) {
      height += (int)cmapOffset[1];
      cmapOffset[1] = 0;
    }

    if (cmapOffset[1]+height >= options->height)
      height = options->height - (int)cmapOffset[1];
  }

  // Create XImage
  if (!(colormapXM = XGetImage(display, pixmap, 0, 0, 
			       width, height, AllPlanes, ZPixmap))) {
    cerr << "FrameTrueColor Internal Error: Unable to Create Colormap XImage" 
	 << endl;
    exit(1);
  }

  // Create Pixmap
  colormapPM = 
    Tk_GetPixmap(display, Tk_WindowId(tkwin), width, height, depth);
  if (!colormapPM) {
    cerr << "FrameTrueColor Internal Error: Unable to Create Colormap Pixmap" 
	 << endl;
    exit(1);
  }

  // colormapGCXOR
  colormapGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);

  // Create table index array
  for (int kk=0; kk<3; kk++) {
    if (colormapData[kk])
      delete [] colormapData[kk];
    colormapData[kk] = new long[width*height];

    if (!colormapData[kk]) {
      cerr << "FrameTrueColor Internal Error: Unable to alloc tmp data array" 
	   << endl;
      exit(1);
    }
  }

  // fill data array
  for (int kk=0; kk<3; kk++) {
    if (!view[kk] || !fits[kk])
      continue;

    if (!(mosaicCount[kk]>1))
      colormapBeginSingle(kk, width, height);
    else {
      if (mosaicFast && mosaicType[kk] == IRAF)
	colormapBeginMosaicFast(kk, width, height);
      else
	colormapBeginMosaic(kk, width, height);
    }
  }
}

void FrameRGBTrueColor::colormapMotionCmd(float rb, float gb, float bb, 
					  float rc, float gc, float bc, int i,
					  unsigned char* cells, int cnt)
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  // first check for change
  if (bias[0] == rb && bias[1] == gb && bias[2] == bb && 
      contrast[0] == rc && contrast[1] == gc && contrast[2] == bc &&
      invert == i && colorCells)
    return;

  // we got a change
  bias[0] = rb;
  bias[1] = gb;
  bias[2] = bb;
  contrast[0] = rc;
  contrast[1] = gc;
  contrast[2] = bc;
  invert = i;

  updateColorCells(cells, cnt);
  updateColorScale();

  // special case
  if ((!view[0] || !fits[0]) &&
      (!view[1] || !fits[1]) &&
      (!view[2] || !fits[2]))
    return;

  int& width = colormapXM->width;
  int& height = colormapXM->height;

  // create img
  unsigned char* img = new unsigned char[width*height*3];
  char* mk = new char[width*height];
  memset(img, 0, width*height*3);
  memset(mk, 0, width*height);

  for (int kk=0; kk<3; kk++) {
    if (!view[kk] || !fits[kk])
      continue;

    const unsigned char* table = colorScale[kk]->psColors();
    long* src = colormapData[kk];
    unsigned char* dest = img;
    char* mkptr = mk;
    for (int jj=0; jj<height; jj++) {
      for (int ii=0; ii<width; ii++, src++, dest+=3, mkptr++)
	if (*src >=0) {
	  memcpy(dest+kk, table+(*src), 1);
	  *mkptr = 1;
	}
    }
  }

  // apply mask
  unsigned char* ptr = img;
  char* mkptr = mk;
  for (int jj=0; jj<height; jj++)
    for (int ii=0; ii<width; ii++, ptr+=3, mkptr++)
      if (!*mkptr) {
	*(ptr  ) = (unsigned char)bgColor->red;
	*(ptr+1) = (unsigned char)bgColor->green;
	*(ptr+2) = (unsigned char)bgColor->blue;
      }

  // clean up
  if (mk)
    delete mk;

  // build colormapXM
  if (img)
    buildColormapXM(img);

  // clean up
  if (img)
    delete img;

  // XImage to Pixmap
  XPutImage(display, colormapPM, Widget::gc, colormapXM, 0, 0, 0, 0, 
	    width, height);

  // Display Pixmap
  Vector d = cmapOffset * widgetToWindow;
  XCopyArea(display, colormapPM, Tk_WindowId(tkwin), colormapGCXOR, 0, 0, 
	    width, height, (int)d[0], (int)d[1]);

  // update panner
  updatePanner();
}

void FrameRGBTrueColor::colormapEndCmd(float rb, float gb, float bb, 
				       float rc, float gc, float bc, int i,
				       unsigned char* cells, int cnt)
{
  if (colormapXM) {
    XDestroyImage(colormapXM);
    colormapXM = NULL;
  }

  if (colormapPM) {
    Tk_FreePixmap(display, colormapPM);
    colormapPM = 0;
  }

  if (colormapGCXOR) {
    XFreeGC(display, colormapGCXOR);
    colormapGCXOR = 0;
  }

  for (int kk=0; kk<3; kk++)
    if (colormapData[kk]) {
      delete [] colormapData[kk];
      colormapData[kk] = NULL;
    }

  update(BASE); // always update
}

void FrameRGBTrueColor::colormapBeginSingle(int kk, int width, int height)
{
  int xoff = (int)cmapOffset[0];
  int yoff = (int)cmapOffset[1];

  double* mm = cfits[kk]->getmWidgetToData();
  int* params = cfits[kk]->getDataParams(frScale[kk].scanMode());
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  int length = colorScale[kk]->size() - 1;

  double ll = cfits[kk]->getLowDouble();
  double hh = cfits[kk]->getHighDouble();
  double diff = hh - ll;

  long* dest = colormapData[kk];
  for (int jj=yoff; jj<yoff+height; jj++) {
    for (int ii=xoff; ii<xoff+width; ii++, dest++) {
      *dest = -1;

      double xx = ii*mm[0] + jj*mm[3] + mm[6];
      double yy = ii*mm[1] + jj*mm[4] + mm[7];

      if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	double value = cfits[kk]->getValueDouble(((int)yy)*srcw + (int)xx);
	int mask =0;
	if (cmask[kk])
	  mask = cmask[kk]->getValueMask(((int)yy)*srcw + (int)xx);
	
	if (!mask) {
	  if (!isnand(value))
	    if (value <= ll)
	      *dest = 0;
	    else if (value >= hh)
	      *dest = length;
	    else
	      *dest = (int)(((value - ll)/diff * length) + .5);
	}
      }
    }
  }
}

void FrameRGBTrueColor::colormapBeginMosaic(int kk, int width, int height)
{
  int xoff = (int)cmapOffset[0];
  int yoff = (int)cmapOffset[1];

  int length = colorScale[kk]->size() - 1;
  long* dest = colormapData[kk];

  for (int jj=yoff; jj<yoff+height; jj++) {
    for (int ii=xoff; ii<xoff+width; ii++, dest++) {
      *dest = -1;

      FitsImage* sptr = cfits[kk];
      FitsImage* mptr = cmask[kk];
      while (sptr) {
	double* mm = sptr->getmWidgetToData();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale[kk].scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  double value = sptr->getValueDouble(((int)yy)*srcw + (int)xx);
	  int mask =0;
	  if (mptr)
	    mask = mptr->getValueMask(((int)yy)*srcw + (int)xx);
	
	  if (!mask) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll)
		*dest = 0;
	      else if (value >= hh)
		*dest = length;
	      else
		*dest = (int)(((value - ll)/diff * length) + .5);
	    }
	  }

	  break;
	}
	else {
	  sptr = sptr->nextMosaic();
	  if (mptr)
	    mptr = mptr->nextMosaic();
	}
      }
    }
  }
}

void FrameRGBTrueColor::colormapBeginMosaicFast(int kk, int width, int height)
{
  int xoff = (int)cmapOffset[0];
  int yoff = (int)cmapOffset[1];

  int length = colorScale[kk]->size() - 1;

  FitsImage* sptr = currentFits;
  FitsImage* mptr = currentMask;
  int found = 0;

  long* dest = colormapData[kk];
  for (int jj=yoff; jj<yoff+height; jj++) {
    for (int ii=xoff; ii<xoff+width; ii++, dest++) {
      *dest = -1;

      while (sptr) {
	double* mm = sptr->getmWidgetToData();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale[kk].scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  found = 1;

	  double value = sptr->getValueDouble(((int)yy)*srcw + (int)xx);
	  int mask =0;
	  if (mptr)
	    mask = mptr->getValueMask(((int)yy)*srcw + (int)xx);
	
	  if (!mask) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll)
		*dest = 0;
	      else if (value >= hh)
		*dest = length;
	      else
		*dest = (int)(((value - ll)/diff * length) + .5);
	    }
	  }

	  break;
	}
	else {
	  if (found) {
	    found = 0;
	    sptr = currentFits;
	    mptr = currentMask;
	  }
	  else {
	    sptr = sptr->nextMosaic();
	    if (mptr)
	      mptr = mptr->nextMosaic();
	  }
	}
      }

      if (!sptr) {
	sptr = currentFits;
	mptr = currentMask;
      }
    }
  }
}

void FrameRGBTrueColor::psLevel2(PSColorSpace mode, 
				 int width, int height, float scale)
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  psLevel2HeadTrue(mode, width, height);

  // special case
  if ((!view[0] || !fits[0]) &&
      (!view[1] || !fits[1]) &&
      (!view[2] || !fits[2]))
    return;

  // update psMatrix
  for (int kk=0; kk<3; kk++) {
    if (fits[kk]) {
      FitsImage* sptr = cfits[kk];
      while (sptr) {
	sptr->updatePS(psMatrix(scale, width, height));
	sptr = sptr->nextMosaic();
      }
    }

    if (mask[kk]) {
      FitsImage* mptr = cmask[kk];
      while (mptr) {
	mptr->updatePS(psMatrix(scale, width, height));
	mptr = mptr->nextMosaic();
      }
    }
  }

  // create img
  unsigned char* img = fillRGBImage(width, height, 0, 0, width, height, 
				    &FitsImage::getmPSToData);
  RLEAscii85 filter;
  unsigned char* ptr = img;

  // size = 2 char per byte * 4 bytes per pixel + estimate of endls + 
  // endl + ends
  int size = (int)(width*4*2*(1 + 1./80.) +.5) + 2;
  char* buf = new char[size];

  for (int jj=0; jj<height; jj++) {
    ostringstream str;

    for (int ii=0; ii<width; ii++, ptr+=3) {
      unsigned char red = ptr[0];
      unsigned char green = ptr[1];
      unsigned char blue = ptr[2];

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
	filter << red << green << blue;
	break;
      case CMYK:
	{
	  unsigned char cyan, magenta, yellow, black;
	  RGB2CMYK(red, green, blue, &cyan, &magenta, &yellow, &black);
	  filter << cyan << magenta << yellow << black;
	}
	break;
      }
      str << filter;
    }
    str << ends;
    psFix(str);
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  ostringstream str;
  filter.flush(str);
  str << endl << "~>" << endl << ends;
  psFix(str);
  Tcl_AppendResult(interp, str.str().c_str(), NULL);

  if (img)
    delete img;
}
