// Copyright (C) 1999-2006
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "tcl.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "frametruecolor.h"
#include "fitsimage.h"
#include "util.h"
#include "ps.h"
#include "NaN.h"

FrameTrueColor::FrameTrueColor(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item)
  : FrameBase(i,c,item), Frame(i,c,item), FrameTrue(i,c,item)
{
  colormapData = NULL;
}

FrameTrueColor::~FrameTrueColor()
{
  if (colormapData)
    delete [] colormapData;
}

void FrameTrueColor::fillXImageSingle(XImage* xmap,
				      int x0, int y0, int x1, int y1,
				      double* (FitsImage::*getMatrix)())
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  if (!currentFits)
    return;

  int& height = xmap->height;
  int bytesPerPixel = xmap->bits_per_pixel/8;

  double* mm = (currentFits->*getMatrix)();
  int* params = currentFits->getDataParams(frScale.scanMode());
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  int length = colorScale->size() - 1;
  int last = length * bytesPerPixel;
  const unsigned char* table = colorScale->colors();

  double ll = currentFits->getLowDouble();
  double hh = currentFits->getHighDouble();
  double diff = hh - ll;

  for (int jj=y0; jj<y1; jj++) {
    // the line may be padded at the end
    char* dest = xmap->data + jj*xmap->bytes_per_line + x0*bytesPerPixel;

    for (int ii=x0; ii<x1; ii++, dest+=bytesPerPixel) {

      double xx = ii*mm[0] + jj*mm[3] + mm[6];
      double yy = ii*mm[1] + jj*mm[4] + mm[7];

      if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	double value = currentFits->getValueDouble(((int)yy)*srcw + (int)xx);
	int mask =0;
	if (currentMask)
	  mask = currentMask->getValueMask(((int)yy)*srcw + (int)xx);

	if (!mask) {
	  if (!isnand(value)) {
	    if (value <= ll)
	      memcpy(dest, table, bytesPerPixel);
	    else if (value >= hh)
	      memcpy(dest, table+last, bytesPerPixel);
	    else
	      memcpy(dest, table+((int)(((value - ll)/diff * length) + .5))*
		     bytesPerPixel,bytesPerPixel);
	  }
	  else
	    memcpy(dest, nanTrueColor_, bytesPerPixel);
	} else
	  memcpy(dest, maskTrueColor_, bytesPerPixel);
      }
      else
	memcpy(dest, bgTrueColor_, bytesPerPixel);
    }
  }
}

void FrameTrueColor::fillXImageMosaic(XImage* xmap, 
				      int x0, int y0, int x1, int y1,
				      double* (FitsImage::*getMatrix)())
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  int& height = xmap->height;
  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->colors();
  int bytesPerPixel = xmap->bits_per_pixel/8;
  int last = length * bytesPerPixel;

  for (int jj=y0; jj<y1; jj++) {

    // the line may be padded at the end
    char* dest = xmap->data + jj*xmap->bytes_per_line + x0*bytesPerPixel;

    for (int ii=x0; ii<x1; ii++, dest+=bytesPerPixel) {
      memcpy(dest, bgTrueColor_, bytesPerPixel);

      FitsImage* sptr = currentFits;
      FitsImage* mptr = currentMask;

      while (sptr) {
	double* mm = (sptr->*getMatrix)();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale.scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  double value = sptr->getValueDouble(((int)yy)*srcw + (int)xx);
	  int mask =0;
	  if (mptr)
	    mask = mptr->getValueMask(((int)yy)*srcw + (int)xx);

	  if (!mask) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll)
		memcpy(dest, table, bytesPerPixel);
	      else if (value >= hh)
		memcpy(dest, table+last, bytesPerPixel);
	      else
		memcpy(dest,
	          table+((int)(((value-ll)/diff*length)+.5))*bytesPerPixel,
		  bytesPerPixel);
	    }
	    else 
	      memcpy(dest, nanTrueColor_, bytesPerPixel);
	  }
	  else
	      memcpy(dest, maskTrueColor_, bytesPerPixel);

	  break;
	}
	else {
	  sptr = sptr->nextMosaic();
	  if (mptr)
	    mptr = mptr->nextMosaic();
	}
      }
    }
  }
}

void FrameTrueColor::fillXImageMosaicFast(XImage* xmap, 
					  int x0, int y0, int x1, int y1,
					  double* (FitsImage::*getMatrix)())
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  if (!currentFits)
    return;

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->colors();

  int& height = xmap->height;
  int bytesPerPixel = xmap->bits_per_pixel/8;
  int last = length * bytesPerPixel;

  FitsImage* sptr = currentFits;
  FitsImage* mptr = currentMask;
  int found = 0;

  for (int jj=y0; jj<y1; jj++) {
    // the line may be padded at the end
    char* dest = xmap->data + jj*xmap->bytes_per_line + x0*bytesPerPixel;

    for (int ii=x0; ii<x1; ii++, dest+=bytesPerPixel) {
      memcpy(dest, bgTrueColor_, bytesPerPixel);

      while (sptr) {
	double* mm = (sptr->*getMatrix)();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale.scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  found = 1;

	  double value = sptr->getValueDouble(((int)yy)*srcw + (int)xx);
	  int mask =0;
	  if (mptr)
	    mask = mptr->getValueMask(((int)yy)*srcw + (int)xx);

	  if (!mask) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll)
		memcpy(dest, table, bytesPerPixel);
	      else if (value >= hh)
		memcpy(dest, table+last, bytesPerPixel);
	      else
		memcpy(dest, table+
		       ((int)(((value-ll)/diff*length)+.5))*bytesPerPixel, 
		       bytesPerPixel);
	    }
	    else 
	      memcpy(dest, nanTrueColor_, bytesPerPixel);
	  }
	  else
	    memcpy(dest, maskTrueColor_, bytesPerPixel);

	  break;
	}
	else {
	  if (found) {
	    found = 0;
	    sptr = currentFits;
	    mptr = currentMask;
	  }
	  else {
	    sptr = sptr->nextMosaic();
	    if (mptr)
	      mptr = mptr->nextMosaic();
	  }
	}
      }

      if (!sptr) {
	sptr = currentFits;
	mptr = currentMask;
      }
    }
  }
}

void FrameTrueColor::colormapCmd(int id, float b, float c, int i, 
				 unsigned short* index, unsigned char* cells, 
				 int cnt)
{
  // first check for change
  if (cmapID == id && bias == b && contrast == c && invert == i && colorCells)
    return;

  // we got a change
  cmapID = id;
  bias = b;
  contrast = c;
  invert = i;

  updateColorCells(index, cells, cnt);
  updateColorScale();

  if (mosaicCount) {
    update(BASE); // always update
    skipUpdate = 0;
  }
  else {
    if (!skipUpdate) {
      update(BASE); // update once
      skipUpdate = 1;
    }      
  }
}

void FrameTrueColor::colormapBeginCmd(const Vector& w)
{
  // we need some fits data
  // we assume the colorScale length will not change during motion calls
  if (!currentFits)
    return;

  int width = options->width;
  int height = options->height;
  cmapOffset = Vector();
  if (cmapArea) {
    width = height = cmapArea;

    switch (cmapMode) {
    case CENTER:
      cmapOffset = (Vector(options->width,options->height) - 
		    Vector(width,height))/2;
      break;
    case CLICK: 
      {
	Vector where = w * canvasToWidget;
	if (where[0] >= 0 && where[0] < options->width &&
	    where[1] >= 0 && where[1] < options->height)
	  cmapOffset = where - Vector(width,height)/2;
	else
	  cmapOffset = (Vector(options->width,options->height) - 
			Vector(width,height))/2;
      }
      break;
    }

    if (cmapOffset[0] < 0) {
      width += (int)cmapOffset[0];
      cmapOffset[0] = 0;
    }
    if (cmapOffset[0]+width >= options->width)
      width = options->width - (int)cmapOffset[0];

    if (cmapOffset[1] < 0) {
      height += (int)cmapOffset[1];
      cmapOffset[1] = 0;
    }

    if (cmapOffset[1]+height >= options->height)
      height = options->height - (int)cmapOffset[1];
  }

  // Create XImage
  if (!(colormapXM = XGetImage(display, pixmap, 0, 0, 
			       width, height, AllPlanes, ZPixmap))) {
    cerr << "FrameTrueColor Internal Error: Unable to Create Colormap XImage" 
	 << endl;
    exit(1);
  }

  // Create Pixmap
  colormapPM = 
    Tk_GetPixmap(display, Tk_WindowId(tkwin), width, height, depth);
  if (!colormapPM) {
    cerr << "FrameTrueColor Internal Error: Unable to Create Colormap Pixmap" 
	 << endl;
    exit(1);
  }

  // colormapGCXOR
  colormapGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);

  // Create table index array
  if (colormapData)
    delete [] colormapData;
  colormapData = new long[width*height];
  if (!colormapData) {
    cerr << "FrameTrueColor Internal Error: Unable to alloc tmp data array" 
	 << endl;
    exit(1);
  }

  // fill data array
  if (!isMosaic()) 
    colormapBeginSingle(width, height);
  else {
    if (mosaicFast && *currentMosaic == IRAF)
      colormapBeginMosaicFast(width, height);
    else
      colormapBeginMosaic(width, height);
  }
}

void FrameTrueColor::colormapBeginSingle(int width, int height)
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  int xoff = (int)cmapOffset[0];
  int yoff = (int)cmapOffset[1];

  int bytesPerPixel = colormapXM->bits_per_pixel/8;

  double* mm = currentFits->getmWidgetToData();
  int* params = currentFits->getDataParams(frScale.scanMode());
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  int length = colorScale->size() - 1;
  int last = length * bytesPerPixel;

  double ll = currentFits->getLowDouble();
  double hh = currentFits->getHighDouble();
  double diff = hh - ll;

  long* dest = colormapData;
  for (int jj=yoff; jj<yoff+height; jj++) {
    for (int ii=xoff; ii<xoff+width; ii++, dest++) {
      *dest = -3; // initialize

      double xx = ii*mm[0] + jj*mm[3] + mm[6];
      double yy = ii*mm[1] + jj*mm[4] + mm[7];

      if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	double value = currentFits->getValueDouble(((int)yy)*srcw + (int)xx);
	int mask =0;
	if (currentMask)
	  mask = currentMask->getValueMask(((int)yy)*srcw + (int)xx);
	  
	if (!mask) {
	  if (!isnand(value)) {
	    if (value <= ll)
	      *dest = 0;
	    else if (value >= hh)
	      *dest = last;
	    else
	      *dest = (int)(((value - ll)/diff * length) + .5)*bytesPerPixel;
	  }
	  else
	    *dest = -1;
	}
	else
	  *dest = -2;
      }
    }
  }
}

void FrameTrueColor::colormapBeginMosaic(int width, int height)
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  int xoff = (int)cmapOffset[0];
  int yoff = (int)cmapOffset[1];

  int bytesPerPixel = colormapXM->bits_per_pixel/8;

  int length = colorScale->size() - 1;
  int last = length * bytesPerPixel;

  // scan fits

  long* dest = colormapData;
  for (int jj=yoff; jj<yoff+height; jj++) {
    for (int ii=xoff; ii<xoff+width; ii++, dest++) {
      *dest = -3;

      FitsImage* sptr = currentFits;
      FitsImage* mptr = currentMask;

      while (sptr) {
	double* mm = sptr->getmWidgetToData();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale.scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  double value = sptr->getValueDouble(((int)yy)*srcw + (int)xx);
	  int mask =0;
	  if (mptr)
	    mask = mptr->getValueMask(((int)yy)*srcw + (int)xx);

	  if (!mask) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll)
		*dest = 0;
	      else if (value >= hh)
		*dest = last;
	      else
		*dest = (int)(((value - ll)/diff * length) + .5)*bytesPerPixel;
	    }
	    else
	      *dest = -1;
	  }
	  else
	    *dest = -2;

	  break;
	}
	else {
	  sptr = sptr->nextMosaic();
	  if (mptr)
	    mptr = mptr->nextMosaic();
	}
      }
    }
  }
}

void FrameTrueColor::colormapBeginMosaicFast(int width, int height)
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  int xoff = (int)cmapOffset[0];
  int yoff = (int)cmapOffset[1];

  int length = colorScale->size() - 1;

  int bytesPerPixel = colormapXM->bits_per_pixel/8;
  int last = length * bytesPerPixel;

  FitsImage* sptr = currentFits;
  FitsImage* mptr = currentMask;
  int found = 0;

  // scan fits
  long* dest = colormapData;
  for (int jj=yoff; jj<yoff+height; jj++) {
    for (int ii=xoff; ii<xoff+width; ii++, dest++) {
      *dest = -3; // initialize

      while (sptr) {
	double* mm = sptr->getmWidgetToData();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale.scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  found = 1;

	  double value = sptr->getValueDouble(((int)yy)*srcw + (int)xx);
	  int mask =0;
	  if (mptr)
	    mask = mptr->getValueMask(((int)yy)*srcw + (int)xx);

	  if (!mask) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll)
		*dest = 0;
	      else if (value >= hh)
		*dest = last;
	      else
		*dest = 
		  (int)(((value - ll)/diff * length) + .5)*bytesPerPixel;
	    }
	    else
	      *dest = -1;
	  }
	  else
	    *dest = -2;

	  break;
	}
	else {
	  if (found) {
	    found = 0;
	    sptr = currentFits;
	    mptr = currentMask;
	  }
	  else {
	    sptr = sptr->nextMosaic();
	    if (mptr)
	      mptr = mptr->nextMosaic();
	  }
	}
      }

      if (!sptr) {
	sptr = currentFits;
	mptr = currentMask;
      }
    }
  }
}

void FrameTrueColor::colormapMotionCmd(int id, float b, float c, int i, 
				       unsigned short* index, 
				       unsigned char* cells, int cnt)
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  // first check for change
  if (cmapID == id && bias == b && contrast == c && invert == i && colorCells)
    return;

  // we got a change
  cmapID = id;
  bias = b;
  contrast = c;
  invert = i;

  updateColorCells(index, cells, cnt);
  updateColorScale();

  // if we have no data, stop now
  if (!currentFits) 
    return;

  // clear ximage
  int& width = colormapXM->width;
  int& height = colormapXM->height;

  int bytesPerPixel = colormapXM->bits_per_pixel/8;
  int& bytesPerLine = colormapXM->bytes_per_line;

  const unsigned char* table = colorScale->colors();

  long* src = colormapData;
  for (int jj=0; jj<height; jj++) {
    // line may be padded at end
    char* dest = colormapXM->data + jj*bytesPerLine;

    for (int ii=0; ii<width; ii++, src++, dest+=bytesPerPixel)
      switch (*src) {
      case -1:
	memcpy(dest, nanTrueColor_, bytesPerPixel);
	break;
      case -2:
	memcpy(dest, maskTrueColor_, bytesPerPixel);
	break;
      case -3:
	memcpy(dest, bgTrueColor_, bytesPerPixel);
	break;
      default:
	memcpy(dest, table+(*src), bytesPerPixel);
	break;
      }
  }

  // XImage to Pixmap
  XPutImage(display, colormapPM, gc, colormapXM, 0, 0, 0, 0, width, height);

  // Display Pixmap
  Vector d = cmapOffset * widgetToWindow;
  XCopyArea(display, colormapPM, Tk_WindowId(tkwin), colormapGCXOR, 0, 0, 
	    width, height, (int)d[0], (int)d[1]);

  // update panner
  updatePanner();
}

void FrameTrueColor::colormapEndCmd(int id, float b, float c, int i, 
				    unsigned short* index, 
				    unsigned char* cells, int cnt)
{
  if (colormapXM) {
    XDestroyImage(colormapXM);
    colormapXM = NULL;
  }

  if (colormapPM) {
    Tk_FreePixmap(display, colormapPM);
    colormapPM = 0;
  }

  if (colormapGCXOR) {
    XFreeGC(display, colormapGCXOR);
    colormapGCXOR = 0;
  }

  if (colormapData) {
    delete [] colormapData;
    colormapData = NULL;
  }

  update(BASE); // always update
}

void FrameTrueColor::psLevel2(PSColorSpace mode, int width, int height, 
			      float scale)
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  psLevel2HeadTrue(mode, width, height);

  if (!isMosaic())
    psLevel2Single(mode, width, height, scale);
  else {
    if (mosaicFast && *currentMosaic == IRAF)
      psLevel2MosaicFast(mode, width, height, scale);
    else
      psLevel2Mosaic(mode, width, height, scale);
  }
}

void FrameTrueColor::psLevel2Single(PSColorSpace mode,
				    int width, int height, float scale)
{
  if (!currentFits)
    return;

  RLEAscii85 filter;
  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  currentFits->updatePS(psMatrix(scale, width, height));

  double* mm = currentFits->getmPSToData();

  int* params = currentFits->getDataParams(frScale.scanMode());
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  double ll = currentFits->getLowDouble();
  double hh = currentFits->getHighDouble();
  double diff = hh - ll;

  for (int jj=0; jj<height; jj++) {
    ostringstream str;

    for (int ii=0; ii<width; ii++) {
      unsigned char red = bgColor->red;
      unsigned char green = bgColor->green;
      unsigned char blue = bgColor->blue;

      double xx = ii*mm[0] + jj*mm[3] + mm[6];
      double yy = ii*mm[1] + jj*mm[4] + mm[7];

      if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	double value = currentFits->getValueDouble(((int)yy)*srcw + (int)xx);
	int mask =0;
	if (currentMask)
	  mask = currentMask->getValueMask(((int)yy)*srcw + (int)xx);
	
	if (!mask) {
	  if (!isnand(value)) {
	    if (value <= ll) {
	      blue = table[0];
	      green = table[1];
	      red = table[2];
	    }
	    else if (value >= hh) {
	      blue = table[length*3];
	      green = table[length*3+1];
	      red = table[length*3+2];
	    }
	    else {
	      int l = (int)(((value - ll)/diff * length) + .5);
	      blue = table[l*3];
	      green = table[l*3+1];
	      red = table[l*3+2];
	    }
	  }
	  else {
	    blue = nanColor->blue;
	    green = nanColor->green;
	    red = nanColor->red;
	  }
	}
	else {
	  blue = maskColor->blue;
	  green = maskColor->green;
	  red = maskColor->red;
	}
      }

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
	filter << red << green << blue;
	break;
      case CMYK:
	{
	  unsigned char cyan, magenta, yellow, black;
	  RGB2CMYK(red, green, blue, &cyan, &magenta, &yellow, &black);
	  filter << cyan << magenta << yellow << black;
	}
	break;
      }
      str << filter;
    }
    str << ends;
    psFix(str);
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  ostringstream str;
  filter.flush(str);
  str << endl << "~>" << endl << ends;
  psFix(str);
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void FrameTrueColor::psLevel2Mosaic(PSColorSpace mode, 
				    int width, int height, float scale)
{
  if (!currentFits)
    return;

  RLEAscii85 filter;
  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  Matrix mx = psMatrix(scale, width, height);

  FitsImage* sptr = currentFits;
  while (sptr) {
    sptr->updatePS(mx);
    sptr = sptr->nextMosaic();
  }

  FitsImage* mptr = currentMask;
  while (mptr) {
    mptr->updatePS(mx);
    mptr = mptr->nextMosaic();
  }

  for (int jj=0; jj<height; jj++) {
    ostringstream str;

    for (int ii=0; ii<width; ii++) {
      unsigned char red = bgColor->red;
      unsigned char green = bgColor->green;
      unsigned char blue = bgColor->blue;

      sptr = currentFits;
      mptr = currentMask;

      while (sptr) {
	double* mm = sptr->getmPSToData();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale.scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  double value = sptr->getValueDouble(((int)yy)*srcw + (int)xx);
	  int mask =0;
	  if (mptr)
	    mask = mptr->getValueMask(((int)yy)*srcw + (int)xx);

	  if (!mask) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll) {
		blue = table[0];
		green = table[1];
		red = table[2];
	      }
	      else if (value >= hh) {
		blue = table[length*3];
		green = table[length*3+1];
		red = table[length*3+2];
	      }
	      else {
		int l = (int)(((value - ll)/diff * length) + .5);
		blue = table[l*3];
		green = table[l*3+1];
		red = table[l*3+2];
	      }
	    }
	    else {
	      blue = nanColor->blue;
	      green = nanColor->green;
	      red = nanColor->red;
	    }
	  }
	  else {
	    blue = maskColor->blue;
	    green = maskColor->green;
	    red = maskColor->red;
	  }

	  break;
	}
	else {
	  sptr = sptr->nextMosaic();
	  if (mptr)
	    mptr = mptr->nextMosaic();
	}
      }

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
	filter << red << green << blue;
	break;
      case CMYK:
	{
	  unsigned char cyan, magenta, yellow, black;
	  RGB2CMYK(red, green, blue, &cyan, &magenta, &yellow, &black);
	  filter << cyan << magenta << yellow << black;
	}
	break;
      }
      str << filter;
    }
    str << ends;
    psFix(str);
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  ostringstream str;
  filter.flush(str);
  str << endl << "~>" << endl << ends;
  psFix(str);
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void FrameTrueColor::psLevel2MosaicFast(PSColorSpace mode, 
					int width, int height, float scale)
{
  if (!currentFits)
    return;

  RLEAscii85 filter;
  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  Matrix mx = psMatrix(scale, width, height);

  FitsImage* sptr = currentFits;
  while (sptr) {
    sptr->updatePS(mx);
    sptr = sptr->nextMosaic();
  }

  FitsImage* mptr = currentMask;
  while (mptr) {
    mptr->updatePS(mx);
    mptr = mptr->nextMosaic();
  }

  sptr = currentFits;
  mptr = currentMask;
  int found =0;

  for (int jj=0; jj<height; jj++) {
    ostringstream str;

    for (int ii=0; ii<width; ii++) {
      unsigned char red = bgColor->red;
      unsigned char green = bgColor->green;
      unsigned char blue = bgColor->blue;

      while (sptr) {
	double* mm = sptr->getmPSToData();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale.scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  found = 1;

	  double value = sptr->getValueDouble(((int)yy)*srcw + (int)xx);
	  int mask =0;
	  if (mptr)
	    mask = mptr->getValueMask(((int)yy)*srcw + (int)xx);

	  if (!mask) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll) {
		blue = table[0];
		green = table[1];
		red = table[2];
	      }
	      else if (value >= hh) {
		blue = table[length*3];
		green = table[length*3+1];
		red = table[length*3+2];
	      }
	      else {
		int l = (int)(((value - ll)/diff * length) + .5);
		blue = table[l*3];
		green = table[l*3+1];
		red = table[l*3+2];
	      }
	    }
	    else {
	      blue = nanColor->blue;
	      green = nanColor->green;
	      red = nanColor->red;
	    }
	  }
	  else {
	    blue = maskColor->blue;
	    green = maskColor->green;
	    red = maskColor->red;
	  }

	  break;
	}
	else {
	  if (found) {
	    found = 0;
	    sptr = currentFits;
	    mptr = currentMask;
	  }
	  else {
	    sptr = sptr->nextMosaic();
	    if (mptr)
	      mptr = mptr->nextMosaic();
	  }
	}
      }

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
	filter << red << green << blue;
	break;
      case CMYK:
	{
	  unsigned char cyan, magenta, yellow, black;
	  RGB2CMYK(red, green, blue, &cyan, &magenta, &yellow, &black);
	  filter << cyan << magenta << yellow << black;
	}
	break;
      }
      str << filter;

      if (!sptr) {
	sptr = currentFits;
	mptr = currentMask;
      }
    }
    str << ends;
    psFix(str);
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  ostringstream str;
  filter.flush(str);
  str << endl << "~>" << endl << ends;
  psFix(str);
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}
